/*******************************************************************************
 * Copyright (c) 2012 Tamas Kleiber 
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * Contributors:
 *     Tamas Kleiber  - initial API and implementation
 *******************************************************************************/
package org.eclipse.cdt.codan.internal.checkers;

import java.util.Iterator;
import java.util.List;
import java.util.regex.Pattern;

import org.eclipse.cdt.codan.checkers.CodanCheckersActivator;
import org.eclipse.cdt.codan.core.cxx.model.AbstractIndexAstChecker;
import org.eclipse.cdt.codan.core.model.ICheckerWithPreferences;
import org.eclipse.cdt.codan.core.model.IProblem;
import org.eclipse.cdt.codan.core.model.IProblemWorkingCopy;
import org.eclipse.cdt.core.dom.ast.ASTVisitor;
import org.eclipse.cdt.core.dom.ast.IASTDeclSpecifier;
import org.eclipse.cdt.core.dom.ast.IASTDeclarator;
import org.eclipse.cdt.core.dom.ast.IASTName;
import org.eclipse.cdt.core.dom.ast.IASTSimpleDeclaration;
import org.eclipse.cdt.core.dom.ast.IASTTranslationUnit;
import org.eclipse.cdt.core.dom.ast.IVariable;

/**
 * This is a naming convention style checker for variables on different scope
 * levels.
 * Pattern parameter is regular expression defining the style.
 */
public class NamingConventionVariableChecker extends AbstractIndexAstChecker implements ICheckerWithPreferences {
	private static final String ER_ID = "org.eclipse.cdt.codan.internal.checkers.NamingConventionVariableChecker"; //$NON-NLS-1$
	public static final String PARAM_GLOBAL_KEY = "pattern_g"; //$NON-NLS-1$
	public static final String PARAM_MODULE_KEY = "pattern_m"; //$NON-NLS-1$
	public static final String PARAM_FUNCSTATIC_KEY = "pattern_s"; //$NON-NLS-1$
	public static final String PARAM_LOCAL_KEY = "pattern_l"; //$NON-NLS-1$
	public static final String PARAM_CHECK_EXTERNS = "param_ext"; //$NON-NLS-1$
	public static final String PARAM_EXCEPT_ARG_LIST = "exceptions"; //$NON-NLS-1$

	public enum Scope {
		GLOBAL, MODULE, FUNCSTATIC, LOCAL, INVALID
	}

	@Override
	public void processAst(IASTTranslationUnit ast) {
		final List<IProblem> pts = getProblemsByMainId(ER_ID, getFile());
		try {
			ast.accept(new ASTVisitor() {
				{
					shouldVisitDeclarators = true;
				}

				@Override
				public int visit(IASTDeclarator declarator) {
					if (declarator.getParent() instanceof IASTSimpleDeclaration) {
						IASTName astName = declarator.getName();
						if (astName.getBinding() instanceof IVariable) {
							String name = astName.toString();
							if (name != null) {
								for (Iterator<IProblem> iterator = pts.iterator(); iterator.hasNext();) {
									String parameter = null;
									String strScope = ""; //$NON-NLS-1$
									IProblem pt = iterator.next();
									Scope scope = getPatternBySymbolScope(pt, (IASTSimpleDeclaration) declarator.getParent());
									switch (scope) {
										case GLOBAL:
											parameter = (String) getPreference(pt, PARAM_GLOBAL_KEY);
											strScope = "global"; //$NON-NLS-1$
											break;
										case MODULE:
											parameter = (String) getPreference(pt, PARAM_MODULE_KEY);
											strScope = "module"; //$NON-NLS-1$
											break;
										case FUNCSTATIC:
											parameter = (String) getPreference(pt, PARAM_FUNCSTATIC_KEY);
											strScope = "function static"; //$NON-NLS-1$
											break;
										case LOCAL:
											parameter = (String) getPreference(pt, PARAM_LOCAL_KEY);
											strScope = "block"; //$NON-NLS-1$
											break;
									}
									if (parameter != null) {
										Pattern pattern = Pattern.compile(parameter);
										if (!pattern.matcher(name).find() && !isFilteredSymbol(name, pt)) {
											reportProblem(pt, astName, strScope, name, parameter);
										}
									}
								}
							}
						}
					}
					return PROCESS_SKIP;
				}
			});
		} catch (Exception e) {
			CodanCheckersActivator.log(e);
		}
	}

	@Override
	public void initPreferences(IProblemWorkingCopy problem) {
		super.initPreferences(problem);
		addPreference(problem, PARAM_GLOBAL_KEY, CheckersMessages.NamingConventionVariableChecker_LabelNamePatternGlobalScope, "^[a-z]"); //$NON-NLS-1$
		addPreference(problem, PARAM_MODULE_KEY, CheckersMessages.NamingConventionVariableChecker_LabelNamePatternModuleScope, "^[a-z]"); //$NON-NLS-1$
		addPreference(problem, PARAM_FUNCSTATIC_KEY, CheckersMessages.NamingConventionVariableChecker_LabelNamePatternFunctionStaticScope,
				"^[a-z]"); //$NON-NLS-1$
		addPreference(problem, PARAM_LOCAL_KEY, CheckersMessages.NamingConventionVariableChecker_LabelNamePatternBlockScope, "^[a-z]"); //$NON-NLS-1$
		addPreference(problem, PARAM_CHECK_EXTERNS, CheckersMessages.NamingConventionVariableChecker_ParameterExternCheck, Boolean.FALSE);
		addListPreference(problem, PARAM_EXCEPT_ARG_LIST, CheckersMessages.GenericParameter_ParameterExceptions,
				CheckersMessages.GenericParameter_ParameterExceptionsItem);
	}

	private boolean shouldReportExternSymbols(IProblem pt) {
		return (Boolean) getPreference(pt, PARAM_CHECK_EXTERNS);
	}

	private boolean isFilteredSymbol(String arg, IProblem pt) {
		return isFilteredArg(arg, pt, PARAM_EXCEPT_ARG_LIST);
	}

	private Scope getPatternBySymbolScope(IProblem pt, IASTSimpleDeclaration simpleDeclaration) {
		Scope scope;
		int storageClass = simpleDeclaration.getDeclSpecifier().getStorageClass();
		// do not report extern symbol name problems if it is not requested
		if (storageClass == IASTDeclSpecifier.sc_extern) {
			if (shouldReportExternSymbols(pt)) {
				scope = Scope.GLOBAL;
			} else {
				scope = Scope.INVALID;
			}
		} else if (storageClass == IASTDeclSpecifier.sc_static) {
			if (isGlobalOrModuleScopeSymbol(simpleDeclaration)) {
				scope = Scope.MODULE;
			} else {
				scope = Scope.FUNCSTATIC;
			}
		} else if (isGlobalOrModuleScopeSymbol(simpleDeclaration)) {
			scope = Scope.GLOBAL;
		} else {
			scope = Scope.LOCAL;
		}
		return scope;
	}

	private boolean isGlobalOrModuleScopeSymbol(IASTSimpleDeclaration simpleDeclaration) {
		return (simpleDeclaration.getParent() instanceof IASTTranslationUnit);
	}

	@Override
	public boolean runInEditor() {
		return true;
	}
}
